/* exec.c
 * Functions for executing a parse tree
 */
#include "sim.h"
#include "server.h"

jmp_buf env;		/* for setjmp/longjmp on errors */

/* error types used for calls to exec_error()
 */
#define et_tmm		1	/* type mismatch */
#define et_div0		2	/* division by 0 */
#define et_nofun	3	/* function not found */
#define et_wrparam	4	/* wrong parameters */
#define et_noplayer	5	/* player doesn't exist */
#define et_noton	6	/* player not logged on */
#define et_nofile	7	/* file not found */
#define et_nonode	8	/* node not found */
#define et_noobject	9	/* object not found */
#define et_noshape	10	/* shape not found */
#define et_nopath	11	/* path not found */
#define et_noperson	12	/* person not found */
#define et_nomethod	13	/* method not found */
#define et_parse	14	/* parse error */
#define et_noattrib	15	/* unknown attribute */
#define et_notlval	16	/* not an lvalue */
#define et_smm		17	/* size mismatch */
#define et_notaperson	18	/* action doesn't belong to a person */
#define et_nourl	19	/* error getting URL */
#define et_execerror	20	/* error executing program */

static variable exec_pnode(action *a, pnode *p);
static int init_var(action *a, variable *v, int st);
static void assign_var(action *a, pnode *p, variable *v, variable val);
static void exec_error(action *a, pnode *p, int errt);
static void check_rf_args(action *a, pnode *p, rfunc fn, variable arg[],
			  int argc);
static variable rfprint(action *a, pnode *p, variable arg[], int argc);
static variable rfsqr(action *a, pnode *p, variable arg[], int argc);
static variable rfcos(action *a, pnode *p, variable arg[], int argc);
static variable rfsin(action *a, pnode *p, variable arg[], int argc);
static variable rfsqrt(action *a, pnode *p, variable arg[], int argc);
static variable rfsend(action *a, pnode *p, variable arg[], int argc);
static variable rfsendstr(action *a, pnode *p, variable arg[], int argc);
static variable rfmoveperson(action *a, pnode *p, variable arg[], int argc);
static variable rfmoveobject(action *a, pnode *p, variable arg[], int argc);
static variable rfrotate(action *a, pnode *p, variable arg[], int argc);
static variable rfmove(action *a, pnode *p, variable arg[], int argc);
static variable rfpexists(action *a, pnode *p, variable arg[], int argc);
static variable rfoexists(action *a, pnode *p, variable arg[], int argc);
static variable rfchpath(action *a, pnode *p, variable arg[], int argc);
static variable rfposition(action *a, pnode *p, variable arg[], int argc);
static variable rfatod(action *a, pnode *p, variable arg[], int argc);
static variable rfdtoa(action *a, pnode *p, variable arg[], int argc);
static variable rfsprint(action *a, pnode *p, variable arg[], int argc);
static variable rfteleport(action *a, pnode *p, variable arg[], int argc);
static variable rfdomethod(action *a, pnode *p, variable arg[], int argc);
static variable rfgetshape(action *a, pnode *p, variable arg[], int argc);
static variable rfputshape(action *a, pnode *p, variable arg[], int argc);
static variable rfsetshattrib(action *a, pnode *p, variable arg[], int argc);
static variable rfgetshattrib(action *a, pnode *p, variable arg[], int argc);
static variable rfgetperson(action *a, pnode *p, variable arg[], int argc);
static variable rfputperson(action *a, pnode *p, variable arg[], int argc);
static variable rfsetprattrib(action *a, pnode *p, variable arg[], int argc);
static variable rfgetprattrib(action *a, pnode *p, variable arg[], int argc);
static variable rfpathsfrom(action *a, pnode *p, variable arg[], int argc);
static variable rfsay(action *a, pnode *p, variable arg[], int argc);
static variable rfyell(action *a, pnode *p, variable arg[], int argc);
static variable rfrotateperson(action *a, pnode *p, variable arg[], int argc);
static variable rfrand(action *a, pnode *p, variable arg[], int argc);
static variable rfsendurl(action *a, pnode *p, variable arg[], int argc);
static variable rfpersonsat(action *a, pnode *p, variable arg[], int argc);
static variable rfsystem(action *a, pnode *p, variable arg[], int argc);
static variable rfmid(action *a, pnode *p, variable arg[], int argc);

static char *data_to_string(void *dp, void (*wfunc)());
int string_to_data(char *data, void *dp, int (*pfunc)());

/* exec_action
 * Execute a method of an object or person
 * a    - action to exec method of
 * p    - player who is doing it  (NULL if triggered by a timer/npc)
 * m    - name of method to exec
 * par  - array of parameters
 * parc - number of parameters
 * Returns 1 if successful, 0 otherwise
 */
int exec_action(action *a, player *p, char *m, char **par, int parc)
{
method *mt = NULL;
char parbuf[256];
int i;
variable tmp;

/* find correct method
 */
for(i=0; i<a->metc; i++) {
	int j;

	for(j=0; j<a->met[i]->namec; j++) {
		if (strequ(a->met[i]->name[j],m)) {
			mt = a->met[i];
			break;
			}
		}
	}
if (!mt) return 0;

/* Set up object,player,actionmethod and params variables
 */
freev(a->var[0]);
freev(a->var[1]);
freev(a->var[2]);
freev(a->var[3]);
a->var[0]->val.str = strdupe(a->name);
a->var[1]->val.str = strdupe(p ? p->name : "");
a->var[2]->val.str = strdupe(m);
a->var[3]->val.arr.n = parc;
a->var[3]->val.arr.a = scalloc(parc,sizeof(variable));
for(i=0; i<parc; i++) {
	variable *av = &a->var[3]->val.arr.a[i];
	av->type = vt_str;
	av->val.str = strdupe(par[i]);
	}

/* Initialise local variables
 */
init_vars(a,mt,vst_local);

/* Execute the method's parse tree
 */
if (setjmp(env) == 0) {
	/* Normal execution
	 */
	exec_pnode(a,mt->ptree);
	}
else {
	/* Jumped here from error
	 */
	logerror("Error executing method %s of %s",m,a->name);
	if (p) server_command(p->fd,com_err,err_execfailed);
	}
return 1;
}


/* exec_pnode
 * Evaluate a node in the parse tree, and return it's value
 */
static variable exec_pnode(action *a, pnode *p)
{
variable rv;

rv.type = vt_void;
rv.name = NULL;
if (p == NULL) {
	return rv;
	}
if (p->type == pn_if) {
	variable cond = exec_pnode(a,p->val.nif.cond);

	if (cond.type != vt_dbl)
		exec_error(a,p,et_tmm);
	if (cond.val.dbl != 0)
		exec_pnode(a,p->val.nif.tr);
	else
		exec_pnode(a,p->val.nif.fl);
	freev(&cond);
	return rv;
	}
else if (p->type == pn_block || p->type == pn_cmds) {
	pnode *stmt;

	for(stmt=p->val.nblock.st; stmt; stmt=stmt->next)
		exec_pnode(a,stmt);
	return rv;
	}
else if (p->type == pn_while) {
	variable cond;

	while(1) {
		cond = exec_pnode(a,p->val.nwhile.cond);
		if (cond.type != vt_dbl)
			exec_error(a,p,et_tmm);
		if (cond.val.dbl == 0)
			break;
		exec_pnode(a,p->val.nwhile.blk);
		freev(&cond);
		}
	return rv;
	}
else if (p->type == pn_do) {
	variable cond;

	while(1) {
		exec_pnode(a,p->val.ndo.blk);
		cond = exec_pnode(a,p->val.ndo.cond);
		if (cond.type != vt_dbl)
			exec_error(a,p,et_tmm);
		if (cond.val.dbl == 0)
			break;
		freev(&cond);
		}
	return rv;
	}
else if (p->type == pn_for) {
	variable init,cond,inc;

	init = exec_pnode(a,p->val.nfor.init);
	freev(&init);
	while(1) {
		cond = exec_pnode(a,p->val.nfor.cond);
		if (cond.type != vt_dbl)
			exec_error(a,p,et_tmm);
		if (cond.val.dbl == 0)
			break;
		exec_pnode(a,p->val.nfor.blk);
		inc = exec_pnode(a,p->val.nfor.inc);
		freev(&inc);
		freev(&cond);
		}
	return rv;
	}
else if (p->type == pn_estat) {
	variable vr = exec_pnode(a,p->val.nestat.exp);

	freev(&vr);
	return rv;
	}
else if (p->type == pn_null) {
	return rv;
	}
else if (p->type == pn_equals || p->type == pn_notequals) {
	variable l = exec_pnode(a,p->val.nbinop.l);
	variable r = exec_pnode(a,p->val.nbinop.r);

	if (l.type != r.type)
		exec_error(a,p,et_tmm);
	rv.type = vt_dbl;
	if (l.type == vt_dbl)
		rv.val.dbl = p->type==pn_equals ? (l.val.dbl == r.val.dbl)
						: (l.val.dbl != r.val.dbl);
	else if (l.type == vt_str)
		rv.val.dbl = p->type==pn_equals ? strequ(l.val.str,r.val.str)
						: !strequ(l.val.str,r.val.str);
	else if (l.type == vt_dblarray) {
		int i;

		if (l.val.arr.n != r.val.arr.n)
			exec_error(a,p,et_smm);
		rv.val.dbl = 1;
		for(i=0; i<l.val.arr.n; i++)
			if (l.val.arr.a[i].val.dbl != r.val.arr.a[i].val.dbl)
				rv.val.dbl = 0;
		if (p->type == pn_notequals)
			rv.val.dbl = !rv.val.dbl;
		}
	else if (l.type == vt_strarray && l.val.arr.n == r.val.arr.n) {
		int i;

		if (l.val.arr.n != r.val.arr.n)
			exec_error(a,p,et_smm);
		rv.val.dbl = 1;
		for(i=0; i<l.val.arr.n; i++)
			if (!strcmp(l.val.arr.a[i].val.str,
				    r.val.arr.a[i].val.str))
				rv.val.dbl = 0;
		if (p->type == pn_notequals)
			rv.val.dbl = !rv.val.dbl;
		}
	freev(&l); freev(&r);
	return rv;
	}
else if (p->type >= pn_add && p->type <= pn_and) {
	variable l = exec_pnode(a,p->val.nbinop.l);
	variable r = exec_pnode(a,p->val.nbinop.r);

	if (l.type != vt_dbl || r.type != vt_dbl)
		exec_error(a,p,et_tmm);
	rv.type = vt_dbl;

	/* normal numeric binary operator
	 */
	switch(p->type) {
	case pn_equals:		rv.val.dbl = (l.val.dbl == r.val.dbl); break;
	case pn_notequals:	rv.val.dbl = (l.val.dbl != r.val.dbl); break;
	case pn_add:		rv.val.dbl = (l.val.dbl + r.val.dbl); break;
	case pn_subtract:	rv.val.dbl = (l.val.dbl - r.val.dbl); break;
	case pn_mult:		rv.val.dbl = (l.val.dbl * r.val.dbl); break;
	case pn_divide:		if (r.val.dbl == 0) exec_error(a,p,et_div0);
				else rv.val.dbl = (l.val.dbl / r.val.dbl);
				break;
	case pn_mod:		rv.val.dbl = ((int)l.val.dbl % (int)r.val.dbl);
				break;
	case pn_greater:	rv.val.dbl = (l.val.dbl > r.val.dbl); break;
	case pn_less:		rv.val.dbl = (l.val.dbl < r.val.dbl); break;
	case pn_grequals:	rv.val.dbl = (l.val.dbl >= r.val.dbl); break;
	case pn_lsequals:	rv.val.dbl = (l.val.dbl <= r.val.dbl); break;
	case pn_or:		rv.val.dbl = (l.val.dbl || r.val.dbl); break;
	case pn_and:		rv.val.dbl = (l.val.dbl && r.val.dbl); break;
		}
	freev(&l);
	freev(&r);
	return rv;
	}
else if (p->type == pn_var) {
	variable ptr = exec_pnode(a,p->val.nvar.var);

	if (ptr.type != vt_var)
		exec_error(a,p,et_tmm);
	rv.type = ptr.val.var->type;
	assign_var(a,p,&rv,*ptr.val.var);
	freev(&ptr);
	return rv;
	}
else if (p->type == pn_assign) {
	variable lv = exec_pnode(a,p->val.nassign.lv);
	variable val = exec_pnode(a,p->val.nassign.val);

	if (lv.type != vt_var)
		exec_error(a,p,et_notlval);
	freev(lv.val.var);
	assign_var(a,p,lv.val.var,val);
	rv = val;
	freev(&lv);
	return rv;
	}
else if (p->type == pn_const) {
	rv.type = vt_dbl;
	rv.val.dbl = p->val.nconst.n;
	return rv;
	}
else if (p->type >= pn_addeq && p->type <= pn_divideeq) {
	variable lv = exec_pnode(a,p->val.nlvalop.lv);
	variable val = exec_pnode(a,p->val.nlvalop.val);

	if (lv.type != vt_var || val.type != vt_dbl ||
	    lv.val.var->type != val.type)
		exec_error(a,p,et_tmm);
	rv.type = lv.val.var->type;
	switch(p->type) {
	case pn_addeq:
		lv.val.var->val.dbl += val.val.dbl; break;
	case pn_subtracteq:
		lv.val.var->val.dbl -= val.val.dbl; break;
	case pn_multeq:
		lv.val.var->val.dbl *= val.val.dbl; break;
	case pn_divideeq:
		lv.val.var->val.dbl /= val.val.dbl; break;
		}
	rv.val.dbl = lv.val.var->val.dbl;
	freev(&lv);
	freev(&val);
	return rv;
	}
else if (p->type >= pn_uminus && p->type <= pn_unot) {
	variable r = exec_pnode(a,p->val.nunop.r);

	if (r.type != vt_dbl)
		exec_error(a,p,et_tmm);
	rv.type = vt_dbl;
	switch(p->type) {
	case pn_uminus:		rv.val.dbl = -r.val.dbl; break;
	case pn_uplus:		rv.val.dbl = +r.val.dbl; break;
	case pn_unot:		rv.val.dbl = !r.val.dbl; break;
		}
	freev(&r);
	return rv;
	}
else if (p->type == pn_concat) {
	variable l = exec_pnode(a,p->val.nconcat.l);
	variable r = exec_pnode(a,p->val.nconcat.r);

	if (l.type != vt_str || r.type != vt_str)
		exec_error(a,p,et_tmm);
	rv.type = vt_str;
	rv.val.str = smalloc(strlen(l.val.str)+strlen(r.val.str)+1);
	sprintf(rv.val.str,"%s%s",l.val.str,r.val.str);
	freev(&l); freev(&r);
	return rv;
	}
else if (p->type == pn_sconst) {
	rv.type = vt_str;
	rv.val.str = strdupe(p->val.nsconst.str);
	return rv;
	}
else if (p->type == pn_lval) {
	rv.type = vt_var;
	rv.val.var = p->val.nlval.loc;
	return rv;
	}
else if (p->type == pn_fcall) {
	rfunc fn = find_runtime_func(p->val.nfcall.name);
	variable arg[20];
	int argc = 0, i;
	pnode *pp;

	if (fn == NULL)
		exec_error(a,p,et_nofun);
	for(pp=p->val.nfcall.ar; pp; pp=pp->next)
		arg[argc++] = exec_pnode(a,pp);
	check_rf_args(a,p,fn,arg,argc);
	rv = fn(a,p,arg,argc);
	for(i=0; i<argc; i++)
		freev(&arg[i]);
	return rv;
	}
else if (p->type == pn_aconst) {
	int i,elemc;
	pnode *pp;
	variable exp;

	for(elemc=0,pp=p->val.naconst.l; pp; pp=pp->next,elemc++)
		;
	if (elemc) {
		rv.val.arr.n = elemc;
		rv.val.arr.a = scalloc(elemc,sizeof(variable));
		for(i=0,pp=p->val.naconst.l; pp; pp=pp->next,i++) {
			rv.val.arr.a[i] = exec_pnode(a,pp);
			}
		for(i=1; i<elemc; i++)
			if (rv.val.arr.a[i].type != rv.val.arr.a[0].type)
				exec_error(a,p,et_tmm);
		if (rv.val.arr.a[0].type == vt_dbl)
			rv.type = vt_dblarray;
		else if (rv.val.arr.a[0].type == vt_str)
			rv.type = vt_strarray;
		else
			exec_error(a,p,et_tmm);
		}
	else {
		rv.val.arr.n = 0;
		rv.val.arr.a = scalloc(1,sizeof(variable));
		rv.type = vt_strarray;
		}
	return rv;
	}
else if (p->type == pn_aindex) {
	variable lv = exec_pnode(a,p->val.naindex.arr);
	variable ind = exec_pnode(a,p->val.naindex.ind);
	int ai;

	if (ind.type != vt_dbl) exec_error(a,p,et_tmm);
	ai = ind.val.dbl;
	if (lv.type == vt_var && (lv.val.var->type == vt_strarray ||
				  lv.val.var->type == vt_dblarray)) {
		/* Getting an lvalue from an lvalue array
		 */
		if (ai < 0 || ai >= lv.val.var->val.arr.n)
			exec_error(a,p,et_smm);
		rv.val.var = &lv.val.var->val.arr.a[ai];
		rv.type = vt_var;
		}
	else if (lv.type == vt_strarray || lv.type == vt_dblarray) {
		/* Getting an expr from an expr array
		 */
		if (ai < 0 || ai >= lv.val.arr.n)
			exec_error(a,p,et_smm);
		rv.type = lv.val.arr.a[ai].type;
		assign_var(a,p,&rv,lv.val.arr.a[ai]);
		}
	else
		exec_error(a,p,et_tmm);
	freev(&ind);
	freev(&lv);
	return rv;
	}
else if (p->type >= pn_preinc && p->type <= pn_postdec) {
	variable v = exec_pnode(a,p->val.nincdec.v);
	variable *lv = v.val.var;

	if (lv->type != vt_dbl) exec_error(a,p,et_tmm);
	rv.type = vt_dbl;
	switch(p->type) {
	case pn_preinc:
		lv->val.dbl += 1; rv.val.dbl = lv->val.dbl; break;
	case pn_postinc:
		rv.val.dbl = lv->val.dbl; lv->val.dbl += 1; break;
	case pn_predec:
		lv->val.dbl -= 1; rv.val.dbl = lv->val.dbl; break;
	case pn_postdec:
		rv.val.dbl = lv->val.dbl; lv->val.dbl -= 1; break;
		}
	freev(&v);
	return rv;
	}
else if (p->type == pn_sizeof) {
	variable so = exec_pnode(a,p->val.nsizeof.a);

	rv.type = vt_dbl;
	if (so.type == vt_strarray || so.type == vt_dblarray)
		rv.val.dbl = so.val.arr.n;
	else if (so.type == vt_str)
		rv.val.dbl = strlen(so.val.str);
	else
		exec_error(a,p,et_tmm);
	freev(&so);
	return rv;
	}
else if (p->type == pn_typeof) {
	variable to = exec_pnode(a,p->val.ntypeof.e);
	char type[256];

	rv.type = vt_str;
	switch(to.type) {
	case vt_void:     strcat(type,"void"); break;
	case vt_dbl:      strcat(type,"dbl"); break;
	case vt_str:      strcat(type,"str"); break;
	case vt_var:      strcat(type,"var"); break;
	case vt_dblarray: sprintf(type,"dbl[%d]",to.val.arr.n); break;
	case vt_strarray: sprintf(type,"str[%d]",to.val.arr.n); break;
		}
	rv.val.str = strdupe(type);
	freev(&to);
	return rv;
	}
else {
	logerror("Unknown parse node type!");
	return rv;
	}
}


/* init_vars
 * Go through all the variables in the action a, and for all who's
 * storage matches st initialise with their init expression. If m isn't NULL,
 * add the additonal condition that only variables in method m are init'ed.
 * For static vars, called just after parsing.
 * For local vars, called before each exec_action
 */
int init_vars(action *a, method *m, int st)
{
int i;
for(i=4; i<a->varc; i++)
	if (!init_var(a,a->var[i],st)) return 0;
for(i=0; i<a->metc; i++)
	if (!m || m == a->met[i]) {
		int j;
		for(j=0; j<a->met[i]->varc; j++)
			if (!init_var(a,a->met[i]->var[j],st)) return 0;
		}
return 1;
}


/* init_var
 * Initialise one variable, if the storage class matches
 */
static int init_var(action *a, variable *v, int st)
{
if (v->storage == st) {
	if (setjmp(env) == 0) {
		freev(v);
		if (v->init) {
			/* Use initializer expression
			 */
			variable val = exec_pnode(a,v->init);

			assign_var(a,v->init,v,val);
			freev(&val);
			}
		else {
			/* Set to some sane default value
			 */
			switch(v->type) {
			case vt_dbl: v->val.dbl = 0; break;
			case vt_str: v->val.str = strdupe(""); break;
			case vt_dblarray:
			case vt_strarray:
				v->val.arr.n = 0;
				v->val.arr.a = scalloc(1,sizeof(variable));
				break;
			case vt_var: v->val.var = NULL; break;
				}
			}
		return 1;
		}
	else
		return 0;
	}
else
	return 1;
}


/* assign_var
 * Copy a value to a variable. Assumes that the target has already been freev()d
 */
static void assign_var(action *a, pnode *p, variable *v, variable val)
{
int i;

if (v->type != val.type) exec_error(a,p,et_tmm);
switch(val.type) {
case vt_dbl:
	v->val.dbl = val.val.dbl;
	break;
case vt_str:
	v->val.str = strdupe(val.val.str);
	break;
case vt_var:
	v->val.var = val.val.var;
	break;
case vt_dblarray:
	v->val.arr.n = val.val.arr.n;
	v->val.arr.a = memdupe(val.val.arr.a,sizeof(variable)*val.val.arr.n);
	break;
case vt_strarray:
	v->val.arr.n = val.val.arr.n;
	v->val.arr.a = memdupe(val.val.arr.a,sizeof(variable)*val.val.arr.n);
	for(i=0; i<v->val.arr.n; i++)
		v->val.arr.a[i].val.str = strdupe(v->val.arr.a[i].val.str);
	break;
	}
}


/* exec_error
 * Print an error message and long-jump out of the execution of this
 * method.
 */
static void exec_error(action *a, pnode *p, int errt)
{
char *estr;
player *pl;

switch(errt) {
case et_tmm: estr = "Type mismatch"; break;
case et_div0: estr = "Division by zero"; break;
case et_nofun: estr = "Function not found"; break;
case et_wrparam: estr = "Incorrect number of parameters"; break;
case et_noplayer: estr = "Player does not exist"; break;
case et_noton: estr = "Player not logged in"; break;
case et_nofile: estr = "File not found"; break;
case et_nonode: estr = "Node does not exist"; break;
case et_noobject: estr = "Object does not exist"; break;
case et_noshape: estr = "Shape does not exist"; break;
case et_nopath: estr = "Path does not exist"; break;
case et_noperson: estr = "Person does not exist"; break;
case et_nomethod: estr = "Method not found"; break;
case et_parse: estr = "Data parse error"; break;
case et_noattrib: estr = "Unknown attribute"; break;
case et_notlval: estr = "Not a lvalue"; break;
case et_smm: estr = "Array size mismatch"; break;
case et_notaperson: estr = "Action is not that of a person"; break;
case et_nourl: estr = "Error accessing URL"; break;
case et_execerror: estr = "Error executing program"; break;
default: estr = "Unknown error!"; break;
	}
logerror("%s : near line %d",estr,p->lineno);
if ((pl = lookup_player(a->var[1]->val.str)))
	server_message(pl->fd,"%s : near line %d",estr,p->lineno);
longjmp(env,-1);
}


/* find_runtime_func
 * Returns a pointer to a named runtime function
 */
rfunc find_runtime_func(char *name)
{
int i;
static struct {
	char *name;
	rfunc f;
	} ftab[] = {	"print",rfprint,"sqr",rfsqr,"cos",rfcos,"sin",rfsin,
			"sqrt",rfsqrt,"send",rfsend,"sendstr",rfsendstr,
			"moveperson",rfmoveperson,"moveobject",rfmoveobject,
			"rotate",rfrotate,"move",rfmove,"pexists",rfpexists,
			"oexists",rfoexists,"chpath",rfchpath,
			"positon",rfposition,"atod",rfatod,"dtoa",rfdtoa,
			"sprint",rfsprint,"teleport",rfteleport,
			"domethod",rfdomethod,"getshape",rfgetshape,
			"putshape",rfputshape,"setshattrib",rfsetshattrib,
			"getshattrib",rfgetshattrib,"getperson",rfgetperson,
			"putperson",rfputperson,"setprattrib",rfsetprattrib,
			"getprattrib",rfgetprattrib,"pathsfrom",rfpathsfrom,
			"say",rfsay,"yell",rfyell,"rotateperson",rfrotateperson,
			"rand",rfrand,"sendurl",rfsendurl,
			"personsat",rfpersonsat,"system",rfsystem,"mid",rfmid
			};

for(i=0; i<sizeof(ftab)/sizeof(ftab[0]); i++)
	if (strequ(ftab[i].name,name))
		return ftab[i].f;
return NULL;
}


/* check_rf_args
 * Check if a run-time function has been passed the right number and type
 * of arguments
 */
static void check_rf_args(action *a, pnode *p, rfunc fn, variable arg[], int argc)
{
int i;
static struct {
	rfunc fp;
	int ac;
	int at[10];
	} rfatab[] = {	{ rfprint,-1 },
			{ rfsqr,1,{vt_dbl} },
			{ rfcos,1,{vt_dbl} },
			{ rfsin,1,{vt_dbl} },
			{ rfsqrt,1,{vt_dbl} },
			{ rfsend,2,{vt_str,vt_str} },
			{ rfsendstr,3,{vt_str,vt_str,vt_str} },
			{ rfmoveperson,2,{vt_str,vt_str} },
			{ rfmoveobject,2,{vt_str,vt_str} },
			{ rfrotate,5,{vt_str,vt_dbl,vt_dbl,vt_dbl,vt_dbl} },
			{ rfmove,4,{vt_str,vt_dbl,vt_dbl,vt_dbl} },
			{ rfpexists,1,{vt_str} },
			{ rfoexists,1,{vt_str} },
			{ rfchpath,3,{vt_str,vt_str,vt_dbl} },
			{ rfposition,1,{vt_str} },
			{ rfatod,1,{vt_str} },
			{ rfdtoa,1,{vt_dbl} },
			{ rfsprint,-1 },
			{ rfteleport,2,{vt_str,vt_str} },
			{ rfdomethod,3,{vt_str,vt_str,vt_strarray} },
			{ rfgetshape,1,{vt_str} },
			{ rfputshape,1,{vt_str} },
			{ rfsetshattrib,-1 },
			{ rfgetshattrib,2,{vt_str,vt_str} },
			{ rfgetperson,1,{vt_str} },
			{ rfputperson,1,{vt_str} },
			{ rfsetprattrib,-1 },
			{ rfgetprattrib,2,{vt_str,vt_str} },
			{ rfpathsfrom,1,{vt_str} },
			{ rfsay,1,{vt_str} },
			{ rfyell,1,{vt_str} },
			{ rfrotateperson,2,{vt_str,vt_str} },
			{ rfrand,1,{vt_dbl} },
			{ rfsendurl,2,{vt_str,vt_str} },
			{ rfpersonsat,1,{vt_str} },
			{ rfsystem,1,{vt_str} },
			{ rfmid,3,{vt_str,vt_dbl,vt_dbl} }
			};

for(i=0; i<sizeof(rfatab)/sizeof(rfatab[0]); i++)
	if (fn == rfatab[i].fp) {
		if (rfatab[i].ac != -1) {
			int j;

			if (rfatab[i].ac != argc)
				exec_error(a,p,et_wrparam);
			for(j=0; j<argc; j++)
				if (rfatab[i].at[j] != arg[j].type)
					exec_error(a,p,et_tmm);
			}
		return;
		}
}


/* rfprint
 * ie. print("hello",7,"wibble");
 */
static variable rfprint(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_void};
player *pl = lookup_player(a->var[1]->val.str);
variable str;

if (pl) {
	str = rfsprint(a,p,arg,argc);
	server_command(pl->fd,com_text,count_chars(str.val.str,'\n')+1);
	fdputs(pl->fd,str.val.str);
	freev(&str);
	}
return rv;
}


/* rfsqr
 * ie. x = sqr(7.0)
 */
static variable rfsqr(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_dbl};

rv.val.dbl = arg[0].val.dbl * arg[0].val.dbl;
return rv;
}


/* rfcos
 * ie. - x = cos(rad);
 */
static variable rfcos(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_dbl};

rv.val.dbl = cos(arg[0].val.dbl);
return rv;
}


/* rfsin
 * ie. y = sin(ang);
 */
static variable rfsin(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_dbl};

rv.val.dbl = sin(arg[0].val.dbl);
return rv;
}


/* rfsqrt
 * ie. - foo = sqrt(64.0);
 */
static variable rfsqrt(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_dbl};

if (arg[0].val.dbl >= 0.0) rv.val.dbl = sqrt(arg[0].val.dbl);
else rv.val.dbl = 0.0;
return rv;
}


/* rfsend
 * ie. - send("foo.gif",player)
 */
static variable rfsend(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_void};
player *pl;
char datafile[256];

if ((pl = lookup_player(arg[1].val.str)) == NULL)
	exec_error(a,p,et_noplayer);
if (pl->fd == -1)
	exec_error(a,p,et_noton);
sprintf(datafile,"%s/object_data/%s/%s",mp->mapdir,a->var[0]->val.str,
					arg[0].val.str);
if (send_mime_data(pl->fd,datafile) == 0)
	exec_error(a,p,et_nofile);
return rv;
}


/* rfsendstr
 * Send mime data to a player
 * ie. - sendstr("http://www.x.org","url",player);
 */
static variable rfsendstr(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_void};
player *pl;

if ((pl = lookup_player(arg[2].val.str)) == NULL)
	exec_error(a,p,et_noplayer);
if (pl->fd == -1)
	exec_error(a,p,et_noton);
send_mime_string(pl->fd,arg[0].val.str,arg[1].val.str);
return rv;
}


/* rfmoveperson
 * Teleport a person to another node in this server or on another server
 * ie. - moveperson(player,"tunnel1");
 */
static variable rfmoveperson(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_void};
person *pr;
node *nd;

if ((pr = lookup_person(arg[0].val.str)) == NULL)
	exec_error(a,p,et_noperson);
if ((nd = lookup_node(arg[1].val.str)) == NULL)
	exec_error(a,p,et_nonode);
person_move(pr,nd);
return rv;
}


/* rfmoveobject
 * Moves an object not in someone's inventory to a new node
 * ie. - moveobject("book","irislab1");
 */
static variable rfmoveobject(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_void};
object *ob;
player *pl;
node *nd;
shape *sh;
double dx, dy, dz;

if ((pl = lookup_player(arg[0].val.str)) == NULL)
	exec_error(a,p,et_noplayer);
if ((ob = lookup_object(arg[0].val.str)) == NULL || ob->inv)
	exec_error(a,p,et_noobject);
ob->pos = nd;
server_command_all(com_moveobject,ob->name,nd->name);
sh = lookup_shape(ob->name);
dx = ob->pos->x - sh->pos.x;
dy = ob->pos->y - sh->pos.y;
dz = ob->pos->z - sh->pos.z;
move_shape(sh,dx,dy,dz);
server_command_all(com_moveshape,sh->name,dx,dy,dz);
return rv;
}


/* rfrotate
 * Set the rotation of a shape about some axis
 * ie. - rotate("foo",90,0,0,1);
 */
static variable rfrotate(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_void};
shape *sh;

if ((sh = lookup_shape(arg[0].val.str)) == NULL)
	exec_error(a,p,et_noshape);
server_command_all(com_rotshape, sh->name, arg[1].val.dbl, arg[2].val.dbl,
		   arg[3].val.dbl, arg[4].val.dbl);
return rv;
}


/* rfmove
 * Move a named 3d shape by some vector
 * ie. - move("gate",0,0,-10);
 */
static variable rfmove(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_void};
shape *sh;

if ((sh = lookup_shape(arg[0].val.str)) == NULL)
	exec_error(a,p,et_noshape);
server_command_all(com_moveshape,sh->name,arg[1].val.dbl,arg[2].val.dbl,
		   arg[3].val.dbl);
move_shape(sh,arg[1].val.dbl,arg[2].val.dbl,arg[3].val.dbl);

return rv;
}


/* rfpexists
 * Returns 1 if a named player exists
 * ie. - print(rfpexists("jcam2"));
 */
static variable rfpexists(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_dbl};

if (lookup_player(arg[0].val.str))
	rv.val.dbl = 1;
else
	rv.val.dbl = 0;
return rv;
}


/* rfoexists
 * Returns 1 if an named object exists
 * ie. - if (oexists("book")) ...
 */
static variable rfoexists(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_dbl};

if (lookup_object(arg[0].val.str))
	rv.val.dbl = 1;
else
	rv.val.dbl = 0;
return rv;
}


/* rfchpath
 * Change the state of a path between 2 nodes (0=open, 1=closed)
 * ie. - chpath("tunnel1","tunnel2",1);
 */
static variable rfchpath(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_void};
node *n1, *n2;
int i;

if (!(n1 = lookup_node(arg[0].val.str)) || !(n2 = lookup_node(arg[1].val.str)))
	exec_error(a,p,et_nonode);
for(i=0; i<n1->linkc; i++)
	if (other_end(n1->link[i],n1) == n2) {
		int cl = arg[2].val.dbl;

		n1->link[i]->closed = cl;
		server_command_all(com_chpath,n1->name,n2->name,cl);
		return rv;
		}
exec_error(a,p,et_nopath);
}


/* rfposition
 * Find the node a person / object is at
 * ie. - print("You are at ",position(player));
 */
static variable rfposition(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_str};
person *pr;
object *ob;

if ((pr = lookup_person(arg[0].val.str))) {
	rv.val.str = strdupe(pr->pos->name);
	return rv;
	}
else if ((ob = lookup_object(arg[0].val.str))) {
	if (ob->pos)
		rv.val.str = strdupe(ob->pos->name);
	else
		rv.val.str = strdupe("");
	return rv;
	}
exec_error(a,p,et_noobject);
}


/* rfatod
 * Convert a string to a double
 * ie. - foo = atod("10.1");
 */
static variable rfatod(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_dbl};

rv.val.dbl = atof(arg[0].val.str);
return rv;
}


/* rfdtoa
 * Convert a double to a string
 * ie. - bar = "The number is " & dtoa(foo);
 */
static variable rfdtoa(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_str};

rv.val.str = strdupe(ftoa(arg[0].val.dbl));
return rv;
}


/* rfsprint
 * Concatenate several parameters of various types into a string
 * ie. foo = sprint("There are ",x," things here");
 */
static variable rfsprint(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_str};
char *prbuf = scalloc(1,256);
int   blen  = 256, bpos = 0;
int i;

/* Collect all output (BUG - array output can go out of bounds)
 */
for(i=0; i<argc; i++) {
	char *addstr;
	char arrbuf[256];
	int j,abp=0;

	if (arg[i].type == vt_void)
		addstr = "(void)";
	else if (arg[i].type == vt_dbl)
		addstr = ftoa(arg[i].val.dbl);
	else if (arg[i].type == vt_str)
		addstr = arg[i].val.str;
	else if (arg[i].type == vt_var)
		addstr = "(var)";
	else if (arg[i].type == vt_strarray) {
		strcpy(arrbuf,"[ ");
		for(j=0; j<arg[i].val.arr.n; j++) {
			abp += strlen(arrbuf+abp);
			sprintf(arrbuf+abp,"\"%s\" %s",
				arg[i].val.arr.a[j].val.str,
				j==arg[i].val.arr.n-1 ? "" : "; ");
			}
		strcat(arrbuf,"]");
		addstr = arrbuf;
		}
	else if (arg[i].type == vt_dblarray) {
		strcpy(arrbuf,"[ ");
		for(j=0; j<arg[i].val.arr.n; j++) {
			abp += strlen(arrbuf+abp);
			sprintf(arrbuf+abp,"%f %s",
				arg[i].val.arr.a[j].val.dbl,
				j==arg[i].val.arr.n-1 ? "" : "; ");
			}
		strcat(arrbuf,"]");
		addstr = arrbuf;
		}
	else
		addstr = "Unknown type";
	while(bpos + strlen(addstr) >= blen) {
		blen *= 2;
		prbuf = srealloc(prbuf,blen);	
		}
	strcpy(&prbuf[bpos],addstr);
	bpos += strlen(addstr);
	}

rv.val.str = prbuf;
return rv;
}


/* rfteleport
 * Teleport a person to a new location
 * ie. - teleport(player,"tunnel13");
 */
static variable rfteleport(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_void};
person *pr = lookup_person(arg[0].val.str);
node *nd   = lookup_node(arg[1].val.str);
player *pl = lookup_player(pr->name), *pp;
double dx,dy,dz;

if (!pr) exec_error(a,p,et_noperson);
if (!nd) exec_error(a,p,et_nonode);

if (pl && pl->fd != -1)
	server_command(pl->fd,com_teleport,nd->name);
server_command_all(com_moveperson,pr->name,nd->name);
dx = nd->x - pr->pos->x;
dy = nd->y - pr->pos->y;
dz = nd->z - pr->pos->z;
move_shape(lookup_shape(pr->name),dx,dy,dz);
reset_player();
while((pp = next_online_player()))
	if (pp->ilevel != lvl_none)
		server_command(pp->fd,com_moveshape,pr->name,dx,dy,dz);
pr->pos = nd;
return rv;
}


/* rfdomethod
 * Invoke a method of another object
 * ie. - domethod("open","dungeon_door",["with";"key"]);
 */
static variable rfdomethod(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_void};
action *ac = lookup_action(arg[1].val.str);
player *pl = lookup_player(a->var[1]->val.str);
char *parv[256];
int parc,i;
jmp_buf oldenv;
int execrv;

if (!ac) exec_error(a,p,et_nomethod);
parc = arg[2].val.arr.n;
for(i=0; i<parc; i++)
	parv[i] = arg[2].val.arr.a[i].val.str;
memcpy(&oldenv,&env,sizeof(jmp_buf));	/* copy error jump location */
execrv = exec_action(ac,pl,arg[0].val.str,parv,parc);
memcpy(&env,&oldenv,sizeof(jmp_buf));	/* restore jump buffer */
if (!execrv)
	exec_error(a,p,et_nomethod);
return rv;
}


/* rfgetshape
 * Get a named shape as a string
 * ie. - foo = getshape("foo");
 */
static variable rfgetshape(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_str};
shape *sh = lookup_shape(arg[0].val.str);

if (!sh) exec_error(a,p,et_noshape);
rv.val.str = data_to_string(sh,write_one_shape);
return rv;
}


/* rfputshape
 * Create a shape from a string
 * ie. putshape("box { corners 0 0 0  1 1 1 }");
 */
static variable rfputshape(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_void};
shape *nsh, *sh;
player *pp;

if (!string_to_data(arg[0].val.str,&nsh,parse_shape))
	exec_error(a,p,et_parse);
/* remove old shape on clients and server
 */
if ((sh = lookup_shape(nsh->name))) {
	server_command_all(com_delshape,sh->name);
	remove_shape(sh);
	free_shape(sh);
	}
/* add new shape on server and clients 
 */
add_shape(nsh);
reset_player();
while((pp = next_online_player())) {
	if (pp->ilevel != lvl_none) {
		server_command(pp->fd,com_addshape,nsh->name);
		write_shape(pp->fd,0,nsh,pp->ilevel,0);
		}
	}
return rv;
}


/* data used by rfsetshattrib and rfgetshattrib calls
 */
static struct {
	char *an;
	int at;
	} shatab[] = {	{ "name",vt_str }, { "colour.red",vt_dbl },
			{ "colour.green",vt_dbl }, { "colour.blue",vt_dbl },
			{ "pos.x",vt_dbl }, { "pos.y",vt_dbl },
			{ "pos.z",vt_dbl }, { "hidden",vt_dbl },
			{ "rot.a",vt_dbl }, { "rot.x",vt_dbl },
			{ "rot.y",vt_dbl }, { "rot.z",vt_dbl },
			{ "sc.x",vt_dbl }, { "sc.y",vt_dbl },
			{ "sc.z",vt_dbl }, { "surf",vt_str } };

/* rfsetshattrib
 * Sets some attribute of a shape
 * ie. - sh = setshattrib(sh,"pos.x",3,4,5);
 * ie. - sh = setshattrib(sh,"name",player);
 */
static variable rfsetshattrib(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_str};
shape *sh;
int i, attrib = -1, valtype;

/* Check arguments
 */
if (argc != 3 || arg[0].type != vt_str || arg[1].type != vt_str)
	exec_error(a,p,et_tmm);
for(i=0; i<sizeof(shatab)/sizeof(shatab[0]); i++)
	if (strequ(arg[1].val.str,shatab[i].an)) {
		attrib = i; valtype = shatab[i].at;
		break;
		}
if (attrib < 0) exec_error(a,p,et_noattrib);
if (valtype != arg[2].type) exec_error(a,p,et_tmm);

/* Get the shape, change the attribute and return the new shape
 */
if (!string_to_data(arg[0].val.str,&sh,parse_shape))
	exec_error(a,p,et_parse);
switch(attrib) {
case 0:	/* name */
	strcpy(sh->name,arg[2].val.str); break;

case 1:	/* colour.red */
	sh->colour.red = arg[2].val.dbl; break;

case 2:	/* colour.green */
	sh->colour.green = arg[2].val.dbl; break;

case 3:	/* colour.blue */
	sh->colour.blue = arg[2].val.dbl; break;

case 4:	/* pos.x */
	sh->pos.x = arg[2].val.dbl; break;

case 5:	/* pos.y */
	sh->pos.y = arg[2].val.dbl; break;

case 6:	/* pos.z */
	sh->pos.z = arg[2].val.dbl; break;

case 7: /* hidden */
	sh->hidden = arg[2].val.dbl; break;

case 8:	/* rot.a */
	sh->rota = arg[2].val.dbl; break;

case 9:	/* rot.x */
	sh->rotx = arg[2].val.dbl; break;

case 10:/* rot.y */
	sh->roty = arg[2].val.dbl; break;

case 11:/* rot.z */
	sh->rotz = arg[2].val.dbl; break;

case 12:/* sc.x */
	sh->scx = arg[2].val.dbl; break;

case 13:/* sc.y */
	sh->scy = arg[2].val.dbl; break;

case 14:/* sc.z */
	sh->scz = arg[2].val.dbl; break;

case 15:/* surf */
	sh->surf = strequ(arg[2].val.str,"flat") ? sft_flat :
		   strequ(arg[2].val.str,"glass") ? sft_glass :
						    sft_shiny;
	break;

	} /* end switch */

rv.val.str = data_to_string(sh,write_one_shape);
free_shape(sh);
return rv;
}


/* rfgetshattrib
 * Get some attribute from a shape string
 * ie. - xpos = getshattrib(sh,"pos.x");
 */
static variable rfgetshattrib(action *a, pnode *p, variable arg[], int argc)
{
variable rv;
shape *sh;
int i,attrib = -1;

/* Check args
 */
for(i=0; i<sizeof(shatab)/sizeof(shatab[0]); i++)
	if (strequ(arg[1].val.str,shatab[i].an))
		attrib = i;
if (attrib < 0) exec_error(a,p,et_noattrib);

/* Get the shape and read out the attrib
 */
if (!string_to_data(arg[0].val.str,&sh,parse_shape))
	exec_error(a,p,et_parse);
switch(attrib) {
case 0:	/* name */
	rv.val.str = strdupe(sh->name); rv.type = vt_str; break;

case 1:	/* colour.red */
	rv.val.dbl = sh->colour.red; rv.type = vt_dbl; break;

case 2:	/* colour.green */
	rv.val.dbl = sh->colour.green; rv.type = vt_dbl; break;

case 3:	/* colour.blue */
	rv.val.dbl = sh->colour.blue; rv.type = vt_dbl; break;

case 4:	/* pos.x */
	rv.val.dbl = sh->pos.x; rv.type = vt_dbl; break;

case 5:	/* pos.y */
	rv.val.dbl = sh->pos.y; rv.type = vt_dbl; break;

case 6:	/* pos.z */
	rv.val.dbl = sh->pos.z; rv.type = vt_dbl; break;

case 7: /* hidden */
	rv.val.dbl = sh->hidden; rv.type = vt_dbl; break;

case 8:	/* rot.a */
	rv.val.dbl = sh->rota; rv.type = vt_dbl; break;

case 9:	/* rot.x */
	rv.val.dbl = sh->rotx; rv.type = vt_dbl; break;

case 10:/* rot.y */
	rv.val.dbl = sh->roty; rv.type = vt_dbl; break;

case 11:/* rot.z */
	rv.val.dbl = sh->rotz; rv.type = vt_dbl; break;

case 12:/* sc.x */
	rv.val.dbl = sh->scx; rv.type = vt_dbl; break;

case 13:/* sc.y */
	rv.val.dbl = sh->scy; rv.type = vt_dbl; break;

case 14:/* sc.z */
	rv.val.dbl = sh->scz; rv.type = vt_dbl; break;

case 15:/* surf */
	rv.val.str = sh->surf == sft_flat ? strdupe("flat") :
		     sh->surf == sft_glass ? strdupe("glass") :
					     strdupe("shiny");
	rv.type = vt_str; break;

	} /* end switch */

free_shape(sh);
return rv;
}


/* rfgetperson
 * Get a names person as a string
 * ie. - myperson = getperson(name);
 */
static variable rfgetperson(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_str};
person *pr = lookup_person(arg[0].val.str);

if (!pr) exec_error(a,p,et_noperson);
rv.val.str = data_to_string(pr,write_person);
return rv;
}


/* rfputperson
 * Create a person from a string
 * ie. - putperson("person { name foo  pos tunnel1 }");
 */
static variable rfputperson(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_void};
person *npr, *pr;
player *pp;

if (!string_to_data(arg[0].val.str,&npr,parse_person))
	exec_error(a,p,et_parse);
/* remove old person on clients and server
 */
if ((pr = lookup_person(npr->name))) {
	server_command_all(com_delperson,pr->name);
	remove_person(pr);
	free_person(pr);
	}
/* add new person on server and clients
 */
add_person(npr);
reset_player();
while((pp = next_online_player())) {
	server_command(pp->fd,com_addperson,npr->name);
	write_person(pp->fd,0,npr);
	}
return rv;
}


/* data used by rfsetprattrib and rfgetprattib calls
 */
static struct {
	char *an;
	int at;
	} pratab[] = {	{ "name",vt_str }, { "pos",vt_str },
			{ "ang",vt_dbl }, { "email",vt_str },
			{ "fullname",vt_str } };


/* rfsetprattrib
 * Sets some attribute of a person from a string
 * ie. - pr = setprattrib(pr,"pos","tunnel3");
 * ie. - pr = setprattrib(pr,"name",newname);
 */
static variable rfsetprattrib(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_str};
person *pr;
int i, attrib=-1, valtype;

/* Check arguments
 */
if (argc != 3 || arg[0].type != vt_str || arg[1].type != vt_str)
	exec_error(a,p,et_tmm);
for(i=0; i<sizeof(pratab)/sizeof(pratab[0]); i++)
	if (strequ(arg[1].val.str,pratab[i].an)) {
		attrib = i; valtype = pratab[i].at;
		break;
		}
if (attrib < 0) exec_error(a,p,et_noattrib);
if (valtype != arg[2].type) exec_error(a,p,et_tmm);

/* Get the person, change the attrib and return the new person
 */
if (!string_to_data(arg[0].val.str,&pr,parse_person))
	exec_error(a,p,et_parse);
switch(attrib) {
case 0:	/* name */
	strcpy(pr->name,arg[2].val.str); break;

case 1:	/* pos */
	if ((pr->pos = lookup_node(arg[2].val.str)) == NULL)
		exec_error(a,p,et_nonode);
	break;

case 2:	/* ang */
	pr->ang = arg[2].val.dbl; break;

case 3:	/* email */
	if (pr->email) free(pr->email);
	pr->email = strdupe(arg[2].val.str); break;

case 4:	/* fullname */
	if (pr->fullname) free(pr->fullname);
	pr->fullname = strdupe(arg[2].val.str); break;

	} /* end switch */

rv.val.str = data_to_string(pr,write_one_shape);
free_person(pr);
return rv;
}


/* rfgetprattrib
 * Gets some attribute of a person from a string
 * ie. - myname = getprattrib(getperson(name),"name");
 * ie. - say("I am at ",getprattrib(myperson,"pos");
 */
static variable rfgetprattrib(action *a, pnode *p, variable arg[], int argc)
{
variable rv;
person *pr;
int i,attrib = -1;

/* Check args
 */
for(i=0; i<sizeof(pratab)/sizeof(pratab[0]); i++)
	if (strequ(arg[1].val.str,pratab[i].an))
		attrib = i;
if (attrib < 0) exec_error(a,p,et_noattrib);

/* Get the person and read out the attrib
 */
if (!string_to_data(arg[0].val.str,&pr,parse_person))
	exec_error(a,p,et_parse);
switch(attrib) {
case 0:	/* name */
	rv.val.str = strdupe(pr->name); rv.type = vt_str; break;

case 1:	/* pos */
	rv.val.str = strdupe(pr->pos->name); rv.type = vt_str; break;

case 2:	/* ang */
	rv.val.dbl = pr->ang; rv.type = vt_dbl; break;

case 3:	/* email */
	rv.val.str = strdupe(pr->email ? pr->email : "");
	rv.type = vt_str; break;

case 4:	/* fullname */
	rv.val.str = strdupe(pr->fullname ? pr->fullname : "");
	rv.type = vt_str; break;
	}

free_person(pr);
return rv;
}


/* rfpathsfrom
 * Get the names of all nodes from a given one, except remote nodes or ones
 * where the path is closed.
 * ie. opts = pathsfrom("tunnel1");
 */
static variable rfpathsfrom(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_strarray};
node *nd = lookup_node(arg[0].val.str);
int i,j;

if (!nd) exec_error(a,p,et_nonode);
rv.val.arr.n = 0;
for(i=0; i<nd->linkc; i++)
	if (!nd->link[i]->closed && !other_end(nd->link[i],nd)->special)
		rv.val.arr.n++;
rv.val.arr.a = scalloc(rv.val.arr.n,sizeof(variable));
for(i=0,j=0; i<nd->linkc; i++)
	if (!nd->link[i]->closed && !other_end(nd->link[i],nd)->special) {
		rv.val.arr.a[j].type = vt_str;
		rv.val.arr.a[j].val.str = strdupe(other_end(nd->link[i],nd)->name);
		j++;
		}
return rv;
}


/* rfsay
 * Send a message to all nearby players
 * ie. - say("hello citizen");
 */
static variable rfsay(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_void};
char **txtv, *msgv[256];
int txtc, i;
player *pp;
person *ppr, *npc;

if ((npc = lookup_person(a->name)) == NULL)
	exec_error(a,p,et_notaperson);
parse_string(arg[0].val.str,&txtv,&txtc);
msgv[0] = a->name;
msgv[1] = ":";
for(i=0; i<txtc; i++)
	msgv[i+2] = txtv[i];
reset_player();
while((pp = next_online_player())) {
	ppr = lookup_person(pp->name);
	if (close_node(ppr->pos,npc->pos))
		server_command(pp->fd,com_msg,msgv,txtc+2);
	}
return rv;
}


/* rfyell
 * Send a message to everyone on this server
 * ie. - yell("hello world");
 */
static variable rfyell(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_void};
char **txtv, *msgv[256];
int txtc, i;
player *pp;
person *npc;

if ((npc = lookup_person(a->name)) == NULL)
	exec_error(a,p,et_notaperson);
parse_string(arg[0].val.str,&txtv,&txtc);
msgv[0] = a->name;
msgv[1] = "yells";
msgv[2] = ":";
for(i=0; i<txtc; i++)
	msgv[i+3] = txtv[i];
server_command_all(com_msg,msgv,txtc+3);
return rv;
}


/* rfrotateperson
 * Rotate a person to face a names node
 * ie. - rotateperson(name,"tunnel6");
 *     - rotateperson(name,pathfrom(getprattrib(me,"pos"))[0]);
 */
static variable rfrotateperson(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_void};
person *pr = lookup_person(arg[0].val.str);
node *nd = lookup_node(arg[1].val.str);
shape *sh;
player *pp;
double ang;

if (!pr) exec_error(a,p,et_noperson);
if (!nd) exec_error(a,p,et_nonode);
ang = rtod(atan2(nd->y - pr->pos->y, nd->x - pr->pos->x));
pr->ang = ang;
sh = lookup_shape(pr->name);
sh->rota = ang; sh->rotx = 0; sh->roty = 0; sh->rotz = 1.0;
reset_player();
while((pp = next_online_player()))
	if (pp->ilevel != lvl_none)
		server_command(pp->fd,com_rotshape,sh->name,ang,0.0,0.0,1.0);
return rv;
}


/* rfrand
 * Generate a random number
 * ie. - x = rand(5);
 */
static variable rfrand(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_dbl};

rv.val.dbl = rand()%((int)arg[0].val.dbl);
return rv;
}


/* rfsendurl
 * Send a URL to a player
 * ie. - sendurl("http://www.cs.monash.edu.au/~jmc/foong2.gif",player);
 */
static variable rfsendurl(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_void};
player *pl;

if ((pl = lookup_player(arg[1].val.str)) == NULL)
	exec_error(a,p,et_noplayer);
if (pl->fd == -1)
	exec_error(a,p,et_noton);
if (!send_url_data(pl->fd,arg[0].val.str))
	exec_error(a,p,et_nourl);
return rv;
}

/* rfpersonsat
 * Get the names of all persons at a given node.
 * ie. - victims = personsat("tunnel5");
 */
static variable rfpersonsat(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_strarray};
node *nd = lookup_node(arg[0].val.str);
person *pr;
int i=0;

if (!nd) exec_error(a,p,et_nonode);
rv.val.arr.n = 0;
reset_person();
while((pr = next_person()))
	if (pr->pos == nd)
		rv.val.arr.n++;
rv.val.arr.a = scalloc(rv.val.arr.n,sizeof(variable));
reset_person();
while((pr = next_person()))
	if (pr->pos == nd) {
		rv.val.arr.a[i].type = vt_str;
		rv.val.arr.a[i].val.str = strdupe(pr->name);
		i++;
		}
return rv;
}


/* rfsystem
 * Pass the given string to /bin/sh for execution and return stdout.
 * ie. - str timenow = system("/bin/date");
 */
static variable rfsystem(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_str};
int tube[2], status;
pid_t pid;

pipe(tube);
if (!(pid = sfork_noreg())) {
	set_signals(SIG_DFL);
	close(0);
	close(1);
	dup2(tube[1],1);
	execlp("/bin/sh","/bin/sh","-c",arg[0].val.str);
	exit(127);
	}
close(tube[1]);
rv.val.str = read_string(tube[0]);
close(tube[0]);
waitpid(pid,&status,0);
if (status)
	exec_error(a,p,et_execerror);
return rv;
}


/* rfmid
 * Extract a substring from a string.
 * ie. - str foo = mid("Hello foo!!",6,3);
 */
static variable rfmid(action *a, pnode *p, variable arg[], int argc)
{
variable rv = {vt_str};
int start = arg[1].val.dbl, len = arg[2].val.dbl;

if (start < 0 || start+len > strlen(arg[0].val.str))
	exec_error(a,p,et_smm);
rv.val.str = smalloc(len+1);
strncpy(rv.val.str,&arg[0].val.str[start],len);
rv.val.str[len] = '\0';
return rv;
}


/* data_to_string
 * Call one of the write_* functions to output into a string
 */
static char *data_to_string(void *dp, void (*wfunc)())
{
int tube[2], status;
pid_t pid;
char *str;

pipe(tube);
if (!(pid = sfork_noreg())) {
	set_signals(SIG_DFL);
	wfunc(tube[1],0,dp);
	exit(0);
	}
close(tube[1]);
str = read_string(tube[0]);
close(tube[0]);
waitpid(pid,&status,0);
return str;
}


/* string_to_data
 * Parse a string as data using one of the parse_ functions. Returns 0
 * on failure, or 1 on success
 */
int string_to_data(char *data, void *dp, int (*pfunc)())
{
int tube[2], status, rv;
pid_t pid;

pipe(tube);
if (!(pid = sfork_noreg())) {
	set_signals(SIG_DFL);
	fdwrite(tube[1],data,1,strlen(data));
	exit(0);
	}
close(tube[1]);
rv = pfunc(tube[0],dp);
close(tube[0]);
waitpid(pid,&status,0);
return rv==0 ? 1 : 0;
}

